@@ -1,4 +1,5 @@
 from __future__ import annotations
+
 # Will no longer be needed at Python 3.10.
 
 import bisect
@@ -12,22 +13,41 @@ import os.path
 import queue
 import string
 import tempfile
+import time
 import threading
 import uuid
-from collections.abc import Sequence, Iterable
 from concurrent.futures import ThreadPoolExecutor, Future
 from contextlib import contextmanager
 from datetime import datetime
+from multiprocessing.util import Finalize
 from pathlib import Path
-from typing import Iterator, Union, List, Dict, Optional, Tuple, Type, Callable
+from typing import (
+    Iterable,
+    Iterator,
+    Union,
+    Sequence,
+    List,
+    Dict,
+    Optional,
+    Tuple,
+    Type,
+    Callable,
+    TypeVar,
+)
 from uuid import uuid4
 
 from upathlib import LocalUpath, Upath  # type: ignore
 from upathlib.serializer import (
-    ByteSerializer, _loads,
-    ZJsonSerializer, ZstdJsonSerializer,
-    PickleSerializer, ZPickleSerializer, ZstdPickleSerializer,
-    OrjsonSerializer, ZOrjsonSerializer, ZstdOrjsonSerializer,
+    ByteSerializer,
+    _loads,
+    ZJsonSerializer,
+    ZstdJsonSerializer,
+    PickleSerializer,
+    ZPickleSerializer,
+    ZstdPickleSerializer,
+    OrjsonSerializer,
+    ZOrjsonSerializer,
+    ZstdOrjsonSerializer,
 )
 
 
@@ -35,33 +55,22 @@ logger = logging.getLogger(__name__)
 
 
 class Dumper:
-    def __init__(self, max_workers: int = 8):
-        assert 0 < max_workers < 20
-        self._max_workers = max_workers
-        self._executor: ThreadPoolExecutor = None  # type: ignore
+    def __init__(self, executor: ThreadPoolExecutor):
+        self._executor: executor = executor
         self._sem: threading.Semaphore = None  # type: ignore
-        # Do not instantiate these now.
-        # They would cause trouble when `Biglist`
-        # file-views are sent to other processes.
-
         self._task_file_data: Dict[Future, tuple] = {}
 
-    def __del__(self):
-        self.cancel()
-
     def _callback(self, t):
         self._sem.release()
         del self._task_file_data[t]
         if t.exception():
             raise t.exception()
 
-    def dump_file(self,
-                  file_dumper: Callable[[Upath, list], None],
-                  data_file: Upath,
-                  data: list):
-        if self._executor is None:
-            self._executor = ThreadPoolExecutor(self._max_workers)
-            self._sem = threading.Semaphore(self._max_workers)
+    def dump_file(
+        self, file_dumper: Callable[[Upath, list], None], data_file: Upath, data: list
+    ):
+        if self._sem is None:
+            self._sem = threading.Semaphore(min(5, self._executor._max_workers))
         self._sem.acquire()  # Wait here if the executor is busy at capacity.
         task = self._executor.submit(file_dumper, data_file, data)
         self._task_file_data[task] = (data_file.name, data)
@@ -83,53 +92,60 @@ class Dumper:
         concurrent.futures.wait(list(self._task_file_data.keys()))
 
     def cancel(self):
-        if self._executor is not None:
-            self._executor.shutdown()
-            self._executor = None
         self._task_file_data = {}
 
 
-class Biglist(Sequence):
-    '''
+T = TypeVar("T")
+
+
+class Biglist(Sequence[T]):
+    """
     Data access is optimized for iteration, whereas random access
     (via index or slice) is less efficient, and assumed to be rare.
-    '''
+    """
+
     registered_storage_formats = {}
 
-    DEFAULT_STORAGE_FORMAT = 'pickle'
+    DEFAULT_STORAGE_FORMAT = "pickle"
 
     @classmethod
-    def register_storage_format(cls,
-                                name: str,
-                                serializer: Type[ByteSerializer],
-                                overwrite: bool = False,
-                                ):
-        good = string.ascii_letters + string.digits + '-_'
+    def register_storage_format(
+        cls,
+        name: str,
+        serializer: Type[ByteSerializer],
+        overwrite: bool = False,
+    ):
+        good = string.ascii_letters + string.digits + "-_"
         assert all(n in good for n in name)
-        if name.replace('_', '-') in cls.registered_storage_formats:
+        if name.replace("_", "-") in cls.registered_storage_formats:
             if not overwrite:
                 raise ValueError(f"serializer '{name}' is already registered")
-        name = name.replace('_', '-')
+        name = name.replace("_", "-")
         cls.registered_storage_formats[name] = serializer
 
     @classmethod
     def get_temp_path(cls) -> Upath:
-        '''Subclass needs to customize this if it prefers to use
+        """Subclass needs to customize this if it prefers to use
         a remote blobstore for temp Biglist.
-        '''
-        path = LocalUpath(os.path.abspath(tempfile.gettempdir()),
-                          str(uuid.uuid4()))  # type: ignore
+        """
+        path = LocalUpath(
+            os.path.abspath(tempfile.gettempdir()), str(uuid.uuid4())
+        )  # type: ignore
         return path  # type: ignore
 
     @classmethod
     def dump_data_file(cls, path: Upath, data: list):
-        serializer = cls.registered_storage_formats[path.suffix.lstrip('.').replace('_', '-')]
+        serializer = cls.registered_storage_formats[
+            path.suffix.lstrip(".").replace("_", "-")
+        ]
         data = [cls.pre_serialize(v) for v in data]
         path.write_bytes(serializer.serialize(data))
 
     @classmethod
-    def load_data_file(cls, path: Upath):
-        deserializer = cls.registered_storage_formats[path.suffix.lstrip('.').replace('_', '-')]
+    def load_data_file(cls, path: Upath) -> List[T]:
+        deserializer = cls.registered_storage_formats[
+            path.suffix.lstrip(".").replace("_", "-")
+        ]
         data = path.read_bytes()
         z = deserializer.deserialize(data)
         return [cls.post_deserialize(v) for v in z]
@@ -145,8 +161,8 @@ class Biglist(Sequence):
             yield
 
     @classmethod
-    def pre_serialize(cls, x):
-        '''When the data element is an instance of a custom type,
+    def pre_serialize(cls, x: T):
+        """When the data element is an instance of a custom type,
         it is preferred to convert it to a native type, such as dict,
         before persisting it to files, especially for long-term storage.
 
@@ -155,24 +171,25 @@ class Biglist(Sequence):
         `pre_serialize` and `post_deserialize`. A good pattern is to define
         instance method `to_dict` and class method `from_dict` on the
         custom class, and call them in `pre_serialize` and `post_deserialize`.
-        '''
+        """
         return x
 
     @classmethod
-    def post_deserialize(cls, x):
-        '''The reverse of `pre_serialize`.'''
+    def post_deserialize(cls, x) -> T:
+        """The reverse of `pre_serialize`."""
         return x
 
     @classmethod
-    def new(cls,
-            path: Union[str, Path, Upath] = None,
-            *,
-            batch_size: int = None,
-            keep_files: bool = None,
-            storage_format: str = None,
-            **kwargs,
-            ):
-        # A Biglist object constrution is in either of the two modes
+    def new(
+        cls,
+        path: Union[str, Path, Upath] = None,
+        *,
+        batch_size: int = None,
+        keep_files: bool = None,
+        storage_format: str = None,
+        **kwargs,
+    ):
+        # A Biglist object construction is in either of the two modes
         # below:
         #    a) create a new Biglist to store new data.
         #    b) create a Biglist object pointing to storage of
@@ -191,7 +208,7 @@ class Biglist(Sequence):
         # both a barebone object that is created in this `new`, and a
         # fleshed out object that already has data.
         #
-        # Some settings may be applicale to an existing Biglist object,
+        # Some settings may be applicable to an existing Biglist object,
         # i.e. they control ways to use the object, and are not an intrinsic
         # property of the object. Hence they can be set to diff values while
         # using an existing Biglist object. Such settings should be
@@ -215,7 +232,7 @@ class Biglist(Sequence):
         elif path.is_file():
             raise FileExistsError(path)
 
-        obj = cls(path, **kwargs)  # type: ignore
+        obj = cls(path, require_exists=False, **kwargs)  # type: ignore
 
         obj.keep_files = keep_files
 
@@ -227,21 +244,30 @@ class Biglist(Sequence):
             # to provide the argument `batch_size`.
         else:
             assert batch_size > 0
-        obj.info['batch_size'] = batch_size
+        obj.info["batch_size"] = batch_size
 
         if storage_format is None:
             storage_format = cls.DEFAULT_STORAGE_FORMAT
-        if storage_format.replace('_', '-') not in cls.registered_storage_formats:
-            raise ValueError(
-                f"invalid value of `storage_format`: '{storage_format}'")
-        obj.info['storage_format'] = storage_format.replace('_', '-')
-        obj.info['storage_version'] = 0
+        if storage_format.replace("_", "-") not in cls.registered_storage_formats:
+            raise ValueError(f"invalid value of `storage_format`: '{storage_format}'")
+        obj.info["storage_format"] = storage_format.replace("_", "-")
+        obj.info["storage_version"] = 1
+        # `storage_version` is a flag for certain breaking changes in the implementation,
+        # such that certain parts of the code (mainly concerning I/O) need to
+        # branch into different treatments according to the version.
+        # This has little relation to `storage_format`.
         # version 0 designator introduced on 2022/3/8
+        # version 1 designator introduced on 2022/7/25
         obj._info_file.write_json(obj.info, overwrite=False)
 
         return obj
 
-    def __init__(self, path: Union[str, Path, Upath]):
+    def __init__(
+        self,
+        path: Union[str, Path, Upath],
+        thread_pool_executor: ThreadPoolExecutor = None,
+        require_exists: bool = True,
+    ):
         if isinstance(path, str):
             path = Path(path)
         if isinstance(path, Path):
@@ -261,42 +287,53 @@ class Biglist(Sequence):
         self._append_buffer: List = []
 
         self.keep_files = True
-        self._file_dumper = Dumper()
+        self._file_dumper = None
+
+        self._thread_pool_ = thread_pool_executor
 
         try:
             # Instantiate a Biglist object pointing to
             # existing data.
             self.info = self._info_file.read_json()
         except FileNotFoundError:
+            if require_exists:
+                raise RuntimeError(
+                    f"Cat not find {self.__class__.__name__} at path '{self.path}'"
+                )
             self.info = {}
 
     @property
     def batch_size(self) -> int:
-        return self.info['batch_size']
+        return self.info["batch_size"]
 
     @property
     def _data_dir(self) -> Upath:
-        return self.path / 'store'
-
-    @property
-    def _data_info_file(self) -> Upath:
-        return self.path / 'datafiles_info.json'
-
-    @property
-    def _info_file(self) -> Upath:
-        return self.path / 'info.json'
-
-    @property
-    def storage_format(self) -> str:
-        return self.info['storage_format'].replace('_', '-')
+        return self.path / "store"
 
     @property
     def datafile_ext(self) -> str:
-        return self.storage_format.replace('-', '_')
+        return self.storage_format.replace("-", "_")
+
+    @property
+    def _info_file(self) -> Upath:
+        return self.path / "info.json"
+
+    @property
+    def storage_format(self) -> str:
+        return self.info["storage_format"].replace("_", "-")
 
     @property
     def storage_version(self) -> int:
-        return self.info.get('storage_version', 0)
+        return self.info.get("storage_version", 0)
+
+    @property
+    def _thread_pool(self):
+        if self._thread_pool_ is None:
+            executor = ThreadPoolExecutor(8)
+            self._thread_pool_ = executor
+            Finalize(self, executor.shutdown)
+
+        return self._thread_pool_
 
     def __bool__(self) -> bool:
         return len(self) > 0
@@ -307,20 +344,24 @@ class Biglist(Sequence):
         else:
             self.destroy()
 
-    def __getitem__(self, idx: int):  # type: ignore
-        '''
+    def __getitem__(self, idx: int) -> T:
+        """
         Element access by single index; negative index works as expected.
 
+        This is not optimized for speed. For example, `self.get_data_files`
+        involves an HTTP call. For better speed, use `__iter__`.
+
         This is called in these cases:
 
             - Access a single item of a Biglist object.
             - Access a single item of a Biglist.view()
             - Access a single item of a slice on a Biglist.view()
             - Iterate a slice on a Biglist.view()
-        '''
+        """
         if not isinstance(idx, int):
             raise TypeError(
-                f'{self.__class__.__name__} indices must be integers, not {type(idx).__name__}')
+                f"{self.__class__.__name__} indices must be integers, not {type(idx).__name__}"
+            )
 
         if idx >= 0 and self._read_buffer_file is not None:
             n1, n2 = self._read_buffer_item_range  # type: ignore
@@ -330,8 +371,13 @@ class Biglist(Sequence):
         if idx < 0 and (-idx) <= len(self._append_buffer):
             return self._append_buffer[idx]
 
-        datafiles = self.get_data_files(lazy=True)
-        length = self._data_files_length_
+        # TODO: this is reliable only if there is no other "worker node"
+        # changing the object by calling `append` or `extend`.
+        datafiles = self.get_data_files()
+        if self._data_files_cumlength_:
+            length = self._data_files_cumlength_[-1]
+        else:
+            length = 0
         idx = range(length + len(self._append_buffer))[idx]
 
         if idx >= length:
@@ -343,15 +389,21 @@ class Biglist(Sequence):
         if self._read_buffer_file is not None:
             n1, n2 = self._read_buffer_item_range  # type: ignore
             if idx < n1:
-                ifile1 = self._read_buffer_file_idx_  # pylint: disable=access-member-before-definition
+                ifile1 = (
+                    self._read_buffer_file_idx_
+                )  # pylint: disable=access-member-before-definition
             elif idx < n2:
                 return self._read_buffer[idx - n1]  # type: ignore
             else:
-                ifile0 = self._read_buffer_file_idx_ + 1  # pylint: disable=access-member-before-definition
+                ifile0 = (
+                    self._read_buffer_file_idx_ + 1
+                )  # pylint: disable=access-member-before-definition
 
         # Now find the data file that contains the target item.
 
-        ifile = bisect.bisect_right(self._data_files_cumlength_, idx, lo=ifile0, hi=ifile1)
+        ifile = bisect.bisect_right(
+            self._data_files_cumlength_, idx, lo=ifile0, hi=ifile1
+        )
         # `ifile`: index of data file that contains the target element.
         # `n`: total length before `ifile`.
         if ifile == 0:
@@ -360,7 +412,10 @@ class Biglist(Sequence):
             n = self._data_files_cumlength_[ifile - 1]
         self._read_buffer_item_range = (n, self._data_files_cumlength_[ifile])
         file = self._data_dir / datafiles[ifile][0]
-        data = self._file_dumper.get_file_data(file)
+        if self._file_dumper is None:
+            data = None
+        else:
+            data = self._file_dumper.get_file_data(file)
         if data is None:
             data = self.load_data_file(file)
         self._read_buffer_file = file
@@ -368,7 +423,7 @@ class Biglist(Sequence):
         self._read_buffer = data
         return data[idx - n]
 
-    def __iter__(self):
+    def __iter__(self) -> Iterator[T]:
         # Assuming the biglist will not change (not being appended to)
         # during iteration.
         self.flush()
@@ -380,27 +435,27 @@ class Biglist(Sequence):
         elif ndatafiles > 1:
             max_workers = min(3, ndatafiles)
             tasks = queue.Queue(max_workers)
-            with ThreadPoolExecutor(max_workers) as executor:
-                for i in range(max_workers):
+            executor = self._thread_pool
+
+            for i in range(max_workers):
+                t = executor.submit(
+                    self.load_data_file, self._data_dir / datafiles[i][0]
+                )
+                tasks.put(t)
+            nfiles_queued = max_workers
+
+            for _ in range(ndatafiles):
+                t = tasks.get()
+                data = t.result()
+
+                if nfiles_queued < ndatafiles:
                     t = executor.submit(
                         self.load_data_file,
-                        self._data_dir / datafiles[i][0]
+                        self._data_dir / datafiles[nfiles_queued][0],
                     )
                     tasks.put(t)
-                nfiles_queued = max_workers
-
-                for _ in range(ndatafiles):
-                    t = tasks.get()
-                    data = t.result()
-
-                    if nfiles_queued < ndatafiles:
-                        t = executor.submit(
-                            self.load_data_file,
-                            self._data_dir / datafiles[nfiles_queued][0]
-                        )
-                        tasks.put(t)
-                        nfiles_queued += 1
-                    yield from data
+                    nfiles_queued += 1
+                yield from data
 
         # I don't think this is necessary.
         if self._append_buffer:
@@ -411,12 +466,14 @@ class Biglist(Sequence):
         # that may be appending to the biglist.
         # In other words, if the current object is one of
         # of a number of workers that are concurrently using
-        # the biglist, then the biglist is not being changed.
-        self.get_data_files(lazy=True)
-        return self._data_files_length_ + len(self._append_buffer)
+        # the biglist, then all the other workers are reading only.
+        self.get_data_files()
+        if self._data_files_cumlength_:
+            return self._data_files_cumlength_[-1] + len(self._append_buffer)
+        return len(self._append_buffer)
 
-    def _append(self, x, concurrent: bool) -> None:
-        '''
+    def append(self, x: T) -> None:
+        """
         Append a single element to the in-memory buffer.
         Once the buffer size reaches `self.batch_size`, the buffer's content
         will be written to a file, and the buffer will re-start empty.
@@ -425,57 +482,52 @@ class Biglist(Sequence):
         its content is not written to disk yet.
         However, at any time, the content of this buffer is included in
         `self.__len__` and in element accesses, including iterations.
-        '''
+        """
         self._append_buffer.append(x)
         if len(self._append_buffer) >= self.batch_size:
-            self._flush(concurrent=concurrent)
-
-    def append(self, x) -> None:
-        self._append(x, concurrent=False)
+            self._flush()
 
     def destroy(self) -> None:
-        '''
+        """
         Clears all the files and releases all in-memory data held by this object,
         so that the object is as if upon `__init__` with an empty directory pointed to
         by `self.path`.
 
         After this method is called, this object is no longer usable.
-        '''
-        self._file_dumper.cancel()
+        """
+        if self._file_dumper is not None:
+            self._file_dumper.cancel()
         self._read_buffer = None
         self._read_buffer_file = None
         self._read_buffer_item_range = None
         self._append_buffer = []
         self.path.rmrf()
 
-    def extend(self, x: Iterable) -> None:
+    def extend(self, x: Iterable[T]) -> None:
         for v in x:
             self.append(v)
 
     def file_view(self, file: Union[Upath, int]) -> FileView:
         if isinstance(file, int):
-            datafiles = self.get_data_files(lazy=True)
+            datafiles = self.get_data_files()
             file = self._data_dir / datafiles[file][0]
-        return FileView(file, self.load_data_file)  # type: ignore
+        return FileView(file, self.__class__.load_data_file)  # type: ignore
 
     def file_views(self) -> List[FileView]:
         # This is intended to facilitate parallel processing,
         # e.g. send views on diff files to diff `multiprocessing.Process`es.
         # However, `iter_files` may be a better way to do that.
         datafiles = self.get_data_files()
-        return [
-            self.file_view(self._data_dir / f)
-            for f, _ in datafiles
-        ]
+        return [self.file_view(self._data_dir / f) for f, _ in datafiles]
 
-    def _flush(self, *, wait: bool = False, concurrent: bool = True):
-        '''
+    def _flush(self, *, wait: bool = False):
+        """
         Persist the content of the in-memory buffer to a file,
         reset the buffer, and update relevant book-keeping variables.
 
         This method is called any time the size of the in-memory buffer
         reaches `self.batch_size`. This happens w/o the user's intervention.
-        '''
+        """
         if not self._append_buffer:
             return
 
@@ -483,30 +535,16 @@ class Biglist(Sequence):
         buffer_len = len(buffer)
         self._append_buffer = []
 
-        def _append_data_file():
-            data_files = self.get_data_files(lazy=not concurrent)
-
-            while True:
-                data_file = f"{uuid4()}.{self.datafile_ext}"
-                # TODO: this could be inefficient if there are many files.
-                if not any((data_file == v[0] for v in data_files)):
-                    break
-
-            data_files.append((data_file, buffer_len))
-            self._data_info_file.write_json(data_files, overwrite=True)
-            self._data_files = data_files
-            return data_file
-
-        if concurrent:
-            with self._lockfile(self._data_info_file.with_suffix('.json.lock')):
-                filename = _append_data_file()
-        else:
-            filename = _append_data_file()
-
-        self._data_files_length_ = self._data_files_length_ + buffer_len
-        self._data_files_cumlength_.append(self._data_files_length_)
+        filename = f"{time.time()}_{uuid4()}_{buffer_len}.{self.datafile_ext}"
+        # File name pattern introduced on 7/25/2022.
+        # this should almost guarantee the file name is unique, hence
+        # we do not need to verify this file name does not exist in `data_files`.
+        # Also include timestamp and item count in the file name, in case
+        # later we decide to use these pieces of info.
 
         data_file = self._data_dir / filename
+        if self._file_dumper is None:
+            self._file_dumper = Dumper(self._thread_pool)
         if wait:
             self._file_dumper.wait()
             self.dump_data_file(data_file, buffer)
@@ -519,11 +557,20 @@ class Biglist(Sequence):
             # data to be accessed property.
 
             # TODO:
-            # what if dump fails later? the index file will be updated
+            # what if dump fails later? The 'n_data_files' file is updated
             # already assuming everything will be fine.
 
+        with self._lockfile(self.path / "_n_datafiles_.txt.lock"):
+            try:
+                n = (self.path / "_n_datafiles_.txt").read_text()
+            except FileNotFoundError:
+                n = 0
+            else:
+                n = int(n)
+            (self.path / "_n_datafiles_.txt").write_text(str(n + 1), overwrite=True)
+
     def flush(self):
-        '''
+        """
         When the user is done adding elements to the list, the buffer size
         may not happen to be `self.batch_size`, hence this method is not called
         automatically,
@@ -534,26 +581,73 @@ class Biglist(Sequence):
 
         In summary, call this method once the user is done with adding elements
         to the list *in this session*, meaning in this run of the program.
-        '''
+        """
         self._flush(wait=True)
 
-    def get_data_files(self, lazy: bool = False) -> list:
-        if lazy and self._data_files is not None:
-            return self._data_files
+    def get_data_files(self) -> list:
+        if self.storage_version < 1:
+            # This may not be totally reliable in every scenario.
+            # The older version had a parameter `lazy`, which is gone now.
+            # After some time we may stop supporting this storage version. (7/27/2022)
+            # However, as long as older datasets are in a "read-only" status,
+            # this is fine.
+            if self._data_files is None:
+                try:
+                    data_info_file = self.path / "datafiles_info.json"
+                    self._data_files = data_info_file.read_json()
+                except FileNotFoundError:
+                    self._data_files = []
+        else:
+            try:
+                nfiles = (self.path / "_n_datafiles_.txt").read_text()
+            except FileNotFoundError:
+                nfiles = 0
+            else:
+                nfiles = int(nfiles)
 
-        try:
-            self._data_files = self._data_info_file.read_json()
-            self._data_files_cumlength_ = list(itertools.accumulate(
-                v[1] for v in self._data_files
-            ))
-            self._data_files_length_ = self._data_files_cumlength_[-1]
-        except FileNotFoundError:
-            self._data_files = []
-            self._data_files_length_ = 0
+            if self._data_files is None or len(self._data_files) != nfiles:
+                files = []
+                if nfiles > 0:
+                    for _ in range(5):
+                        # Starting with storage_version 1, data file name is
+                        #   <timestamp>_<uuid>_<itemcount>.<ext>
+                        # <timestamp> contains a '.', no '_';
+                        # <uuid> contains '-', no '_';
+                        # <itemcount> contains no '-' nor '_';
+                        # <ext> may contain '_'.
+                        files0 = (v.name for v in self._data_dir.iterdir())
+                        files1 = (v.split("_") + [v] for v in files0)
+                        files2 = (
+                            (float(v[0]), v[-1], int(v[2].partition(".")[0]))
+                            for v in files1
+                        )
+                        files = sorted(files2)
+                        if len(files) == nfiles:
+                            break
+                        time.sleep(0.2)
+
+                    if len(files) != nfiles:
+                        raise RuntimeError(
+                            f"{nfiles} data files are expected, yet only {len(files)} are found"
+                        )
+
+                if files:
+                    self._data_files = [
+                        (v[1], v[2]) for v in files
+                    ]  # file name, item count
+                else:
+                    self._data_files = []
+
+        if self._data_files:
+            self._data_files_cumlength_ = list(
+                itertools.accumulate(v[1] for v in self._data_files)
+            )
+        else:
             self._data_files_cumlength_ = []
+
         return self._data_files  # type: ignore
 
-    def view(self) -> ListView:
+    def view(self) -> ListView[T]:
         # During the use of this view, the underlying Biglist should not change.
         # Multiple views may be used to view diff parts
         # of the Biglist; they open and read files independent of
@@ -562,43 +656,44 @@ class Biglist(Sequence):
         return ListView(self.__class__(self.path))
 
     def _concurrent_iter_info_file(self, task_id: str) -> Upath:
-        return self.path / 'concurrent_iter' / task_id / 'iter_info.json'
-
-    def concurrent_append(self, x) -> None:
-        self._append(x, concurrent=True)
-
-    def concurrent_extend(self, x: Iterable) -> None:
-        for v in x:
-            self.concurrent_append(v)
+        """
+        `task_id`: returned by `new_concurrent_iter`.
+        """
+        return self.path / ".concurrent_iter" / task_id / "info.json"
 
     def new_concurrent_iter(self) -> str:
-        '''
+        """
         One worker, such as a "coordinator", calls this method once.
         After that, one or more workers independently call `concurrent_iter`
-        to iterate over the Biglist, providing the iter-ID returned by
+        to iterate over the Biglist. `concurrent_iter` takes the task-ID returned by
         this method. The content of the Biglist is
         split between the workers because each data file will be obtained
         by exactly one worker.
-        '''
+
+        During this iteration, the Biglist object should stay unchanged---no
+        calls to `append` and `extend`.
+        """
         task_id = datetime.utcnow().isoformat()
         self._concurrent_iter_info_file(task_id).write_json(
-            {'n_files_claimed': 0}, overwrite=True)
+            {"n_files_claimed": 0}, overwrite=True
+        )
         return task_id
 
-    def concurrent_iter(self, task_id: str, *, reader_id: str = None) -> Iterator:
-        if not reader_id and isinstance(self.path, LocalUpath):
-            reader_id = multiprocessing.current_process().name
+    def concurrent_iter(self, task_id: str) -> Iterator[T]:
+        """
+        `task_id`: returned by `new_concurrent_iter`.
+        """
         datafiles = self.get_data_files()
         while True:
             ff = self._concurrent_iter_info_file(task_id)
-            with self._lockfile(ff.with_suffix('.json.lock')):
+            with self._lockfile(ff.with_suffix(".json.lock")):
                 iter_info = ff.read_json()
-                n_files_claimed = iter_info['n_files_claimed']
+                n_files_claimed = iter_info["n_files_claimed"]
                 if n_files_claimed >= len(datafiles):
                     # No more date files to process.
                     break
 
-                iter_info['n_files_claimed'] = n_files_claimed + 1
+                iter_info["n_files_claimed"] = n_files_claimed + 1
                 ff.write_json(iter_info, overwrite=True)
                 file_name = datafiles[n_files_claimed][0]
                 fv = self.file_view(self._data_dir / file_name)
@@ -608,20 +703,90 @@ class Biglist(Sequence):
             logger.debug('yielding data of file "%s"', file_name)
             yield from fv.data  # this is the data list contained in the data file
 
-    def concurrent_iter_done(self, task_id: str, lazy: bool = True) -> Optional[bool]:
-        try:
-            iter_info = self._concurrent_iter_info_file(task_id).read_json()
-        except FileNotFoundError:
-            return None
-        datafiles = self.get_data_files(lazy=lazy)
-        # Usually, this method is called by a "controller" node to
-        # check the status of iteration of the biglist by multiple worker nodes.
-        # In the meantime, the biglist remain unchanged, hence
-        # `get_data_files` can use the lazy mode.
-        return iter_info['n_files_claimed'] >= len(datafiles)
+    def concurrent_iter_stat(self, task_id: str) -> dict:
+        info = self._concurrent_iter_info_file(task_id).read_json()
+        return {**info, "n_files": len(self.get_data_files())}
+
+    def concurrent_iter_done(self, task_id: str) -> bool:
+        zz = self.concurrent_iter_stat(task_id)
+        return zz["n_files_claimed"] >= zz["n_files"]
+
+    def _multiplex_info_file(self, task_id: str) -> Upath:
+        """
+        `task_id`: returned by `new_multiplexer`.
+        """
+        return self.path / ".multiplexer" / task_id / "info.json"
+
+    def new_multiplexer(self) -> str:
+        """
+        One worker, such as a "coordinator", calls this method once.
+        After that, one or more workers independently call `multiplex_iter`
+        to iterate over the Biglist. `multiplex_iter` takes the task-ID returned by
+        this method. The content of the Biglist is
+        split between the workers because each data item will be obtained
+        by exactly one worker.
+
+        During this iteration, the Biglist object should stay unchanged---no
+        calls to `append` and `extend`.
+
+        Difference between `concurrent_iter` and `multiplex_iter`: the former
+        distributes files to workers, whereas the latter distributes individual
+        data elements to workers.
+
+        Use case of multiplexer: each data element represents considerable amounts
+        of work--it is a "hyper-parameter" or the like; `multiplex_iter` facilitates
+        splitting the work represented by different values of the "hyper-parameter"
+        between multiple workers.
+        """
+        assert not self._append_buffer
+        task_id = datetime.utcnow().isoformat()
+        self._multiplex_info_file(task_id).write_json(
+            {
+                "total": len(self),
+                "next": 0,
+                "time": datetime.utcnow().isoformat(),
+            },
+            overwrite=False,
+        )
+        return task_id
+
+    def multiplex_iter(self, task_id: str, worker_id: str = None) -> Iterator[T]:
+        """
+        `task_id`: returned by `new_multiplexer`.
+        """
+        if not worker_id:
+            worker_id = "{} {}".format(
+                multiprocessing.current_process().name,
+                threading.current_thread().name,
+            )
+        while True:
+            f = self._multiplex_info_file(task_id)
+            with f.with_suffix(f.suffix + ".lock").lock():
+                ss = self._multiplex_info_file(task_id).read_json()
+                n = ss["next"]
+                if n == ss["total"]:
+                    return
+                self._multiplex_info_file(task_id).write_json(
+                    {
+                        "next": n + 1,
+                        "worker_id": worker_id,
+                        "time": datetime.utcnow().isoformat(),
+                        "total": ss["total"],
+                    },
+                    overwrite=True,
+                )
+            yield self[n]
+
+    def multiplex_stat(self, task_id: str) -> dict:
+
+        return self._multiplex_info_file(task_id).read_json()
+
+    def multiplex_done(self, task_id: str) -> bool:
+        ss = self.multiplex_stat(task_id)
+        return ss["next"] == ss["total"]
 
 
-class FileView(Sequence):
+class FileView(Sequence[T]):
     def __init__(self, file: Upath, loader: Callable):
         # TODO: make `Upath` safe to pass across processes.
         self._file = file
@@ -629,24 +794,24 @@ class FileView(Sequence):
         self._data = None
 
     @property
-    def data(self):
+    def data(self) -> List[T]:
         if self._data is None:
             self._data = self._loader(self._file)
         return self._data
 
-    def __len__(self):
+    def __len__(self) -> int:
         return len(self.data)
 
-    def __getitem__(self, idx: Union[int, slice]):
+    def __getitem__(self, idx: Union[int, slice]) -> T:
         return self.data[idx]
 
     def __iter__(self):
         return iter(self.data)
 
 
-class ListView(Sequence):
-    def __init__(self, list_: Sequence, range_: range = None):
-        '''
+class ListView(Sequence[T]):
+    def __init__(self, list_: Sequence[T], range_: range = None):
+        """
         This provides a "window" into the sequence `list_`,
         which is often a `Biglist` or another `ListView`.
 
@@ -658,7 +823,7 @@ class ListView(Sequence):
 
         During the use of this object, it is assumed that the underlying
         `list_` is not changing. Otherwise the results may be incorrect.
-        '''
+        """
         self._list = list_
         self._range = range_
 
@@ -670,13 +835,13 @@ class ListView(Sequence):
     def __bool__(self) -> bool:
         return len(self) > 0
 
-    def __getitem__(self, idx: Union[int, slice]):
-        '''
+    def __getitem__(self, idx: Union[int, slice]) -> T:
+        """
         Element access by a single index or by slice.
         Negative index and standard slice syntax both work as expected.
 
         Sliced access returns a new `ListView` object.
-        '''
+        """
         if isinstance(idx, int):
             if self._range is None:
                 return self._list[idx]
@@ -690,7 +855,8 @@ class ListView(Sequence):
             return self.__class__(self._list, range_)
 
         raise TypeError(
-            f'{self.__class__.__name__} indices must be integers or slices, not {type(idx).__name__}')
+            f"{self.__class__.__name__} indices must be integers or slices, not {type(idx).__name__}"
+        )
 
     def __iter__(self):
         if self._range is None:
@@ -700,7 +866,7 @@ class ListView(Sequence):
                 yield self._list[i]
 
     @property
-    def raw(self):
+    def raw(self) -> Sequence[T]:
         return self._list
 
 
@@ -714,12 +880,12 @@ class JsonByteSerializer(ByteSerializer):
         return _loads(json.loads, y.decode(), **kwargs)
 
 
-Biglist.register_storage_format('json', JsonByteSerializer)
-Biglist.register_storage_format('json-z', ZJsonSerializer)
-Biglist.register_storage_format('json-zstd', ZstdJsonSerializer)
-Biglist.register_storage_format('pickle', PickleSerializer)
-Biglist.register_storage_format('pickle-z', ZPickleSerializer)
-Biglist.register_storage_format('pickle-zstd', ZstdPickleSerializer)
-Biglist.register_storage_format('orjson', OrjsonSerializer)
-Biglist.register_storage_format('orjson-z', ZOrjsonSerializer)
-Biglist.register_storage_format('orjson-zstd', ZstdOrjsonSerializer)
+Biglist.register_storage_format("json", JsonByteSerializer)
+Biglist.register_storage_format("json-z", ZJsonSerializer)
+Biglist.register_storage_format("json-zstd", ZstdJsonSerializer)
+Biglist.register_storage_format("pickle", PickleSerializer)
+Biglist.register_storage_format("pickle-z", ZPickleSerializer)
+Biglist.register_storage_format("pickle-zstd", ZstdPickleSerializer)
+Biglist.register_storage_format("orjson", OrjsonSerializer)
+Biglist.register_storage_format("orjson-z", ZOrjsonSerializer)
+Biglist.register_storage_format("orjson-zstd", ZstdOrjsonSerializer)
